package mhtree;

import cz.muni.fi.disa.similarityoperators.cover.AbstractRepresentation.PrecomputedDistances;
import cz.muni.fi.disa.similarityoperators.cover.HullOptimizedRepresentationV3;
import messif.buckets.BucketStorageException;
import messif.objects.LocalAbstractObject;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.stream.Collectors;

public abstract class Node implements Serializable {

    /**
     * Serialization ID
     */
    private static final long serialVersionUID = 420L;

    private final InsertType INSERT_TYPE;
    private final ObjectToNodeDistance OBJECT_TO_NODE_DISTANCE;

    private HullOptimizedRepresentationV3 hull;

    protected Node(PrecomputedDistances distances, InsertType insertType, ObjectToNodeDistance objectToNodeDistance) {
        this.hull = new HullOptimizedRepresentationV3(distances);
        this.hull.build();

        this.INSERT_TYPE = insertType;
        this.OBJECT_TO_NODE_DISTANCE = objectToNodeDistance;
    }

    protected static InternalNode createParent(List<Node> nodes, PrecomputedDistances distances, InsertType insertType, ObjectToNodeDistance objectToNodeDistance) {
        List<LocalAbstractObject> objects = nodes
                .stream()
                .map(Node::getObjects)
                .flatMap(Collection::stream)
                .collect(Collectors.toList());

        if (nodes.size() == distances.getObjectCount()) {
            return new InternalNode(distances, insertType, objectToNodeDistance, nodes);
        }

        return new InternalNode(distances.getSubset(objects), insertType, objectToNodeDistance, nodes);
    }

    /**
     * Returns a list of hull objects.
     *
     * @return a list of hull objects
     */
    public List<LocalAbstractObject> getHullObjects() {
        return hull.getHull();
    }

    /**
     * Returns true if the {@code object} is covered.
     *
     * @param object the object to be checked
     * @return true if the {@code object} is covered.
     */
    public boolean isCovered(LocalAbstractObject object) {
        return hull.isExternalCovered(object);
    }

    @Override
    public String toString() {
        return "Node{hull=" + hull + '}';
    }

    protected double getDistance(LocalAbstractObject object) {
        return OBJECT_TO_NODE_DISTANCE.getDistance(object, this);
    }

    protected double getDistance(LocalAbstractObject object, PrecomputedDistances distances) {
        return OBJECT_TO_NODE_DISTANCE.getDistance(object, this, distances);
    }

    protected double getDistanceToNearest(LocalAbstractObject object) {
        return ObjectToNodeDistance.NEAREST.getDistance(object, this);
    }

    protected boolean isLeaf() {
        return (this instanceof LeafNode);
    }

    protected boolean isInternal() {
        return !isLeaf();
    }

    protected int getHullObjectCount() {
        return hull.getRepresentativesCount();
    }

    protected void addObjectIntoHull(LocalAbstractObject object) {
        addObjectIntoHull(object, null);
    }

    protected void addObjectIntoHull(LocalAbstractObject object, PrecomputedDistances distances) {
        if (isCovered(object, distances)) return;

        if (INSERT_TYPE == InsertType.INCREMENTAL) {
            hull.addHullObject(object);
            return;
        }

        rebuildHull(object, distances);
    }

    protected abstract void addObject(LocalAbstractObject object) throws BucketStorageException;

    protected abstract List<LocalAbstractObject> getObjects();

    protected abstract int getHeight();

    protected abstract void gatherNodes(List<Node> nodes);

    protected abstract void gatherLeafNodes(List<LeafNode> leafNodes);

    private void rebuildHull(LocalAbstractObject object, PrecomputedDistances distances) {
        List<LocalAbstractObject> objects = new ArrayList<>(getObjects());
        objects.add(object);

        if (distances == null) {
            hull = new HullOptimizedRepresentationV3(objects);
        } else {
            hull = new HullOptimizedRepresentationV3(distances.getSubset(objects));
        }

        hull.build();
    }

    private boolean isCovered(LocalAbstractObject object, PrecomputedDistances distances) {
        if (distances == null) {
            return hull.isExternalCovered(object);
        }

        return hull.isExternalCovered(object, distances);
    }
}
